// ----------------------------------------------------------------
// The contents of this file are distributed under the CC0 license.
// See http://creativecommons.org/publicdomain/zero/1.0/
// ----------------------------------------------------------------

#include "error.h"
#include "string.h"
#include "json.h"
#include "rand.h"
#include "matrix.h"
#include "neuralnet.h"
#include <stdio.h>
#include <stdlib.h>
#include <iostream>
#include <exception>
#include <string>
#include <memory>
#ifdef WINDOWS
#	include <io.h> // for "access"
#else
#	include <unistd.h>
#endif

using std::cout;
using std::cerr;
using std::string;
using std::vector;

void train_neural_net()
{
	NeuralNet* pNN = nullptr;
	if(access("model.json", 0) == 0) // If the file "model.json" exists...
	{
		// Load the existing model from the file
		cout << "Loading the model from file...\n";
		Json doc;
		doc.loadJson("model.json");
		pNN = new NeuralNet(*doc.root());
	}
	else
	{
		// Make a new model
		cout << "Making a new model...\n";
		pNN = new NeuralNet();
		pNN->layers().push_back(new LayerLinear(3, 10));
		pNN->layers().push_back(new LayerTanh(10));
		pNN->layers().push_back(new LayerLinear(10, 8));
		pNN->layers().push_back(new LayerTanh(8));
		pNN->layers().push_back(new LayerLinear(8, 1));
		pNN->layers().push_back(new LayerTanh(1));
		Rand rand(1234);
		pNN->init_weights(rand);

		// Train the model to map from <0.1, 0.2, 0.3> to <0.4>
		cout << "Training it...\n";
		Vec in(3);
		in[0] = 0.1;
		in[1] = 0.2;
		in[2] = 0.3;
		Vec target(1);
		target[0] = 0.4;
		double learning_rate = 0.03;
		for(size_t i = 0; i < 5000; i++)
		{
			pNN->train_incremental(in, target, learning_rate);
		}		

		// Save the model to a file
		cout << "Saving it to file...\n";
		Json doc;
		JsonNode* pNode = pNN->marshal(doc);
		pNode->saveJson("model.json");
	}
	
	// Test the model
	cout << "Testing the model...\n";
	Vec in(3);
	in[0] = 0.1;
	in[1] = 0.2;
	in[2] = 0.3;
	const Vec& prediction = pNN->forwardprop(in);
	cout << "Prediction: " << to_str(prediction) << "\n";
	cout << "Goodbye.";
	delete(pNN);
}

int main(int argc, char *argv[])
{
	enableFloatingPointExceptions();
	int ret = 1;
	try
	{
		NeuralNet::unit_test1();
		train_neural_net();
		ret = 0;
	}
	catch(const std::exception& e)
	{
		cerr << "An error occurred: " << e.what() << "\n";
	}
	return ret;
}
